/***************************************************************************************/
/*                                                                                     */
/*           RIP protocol : common parameters                                          */
/*                                                                                     */
/***************************************************************************************/
/***************************************************************************************/
/*                                                                                     */
/*           Pierre Rousseau and Carlos Aguilar Melchor.                               */
/*           Limoges University, France.                                               */
/*           All rights reserved.                                                      */
/*           Note: once our work is finished all this work will be under a GPL licence.*/
/*           Until then, we keep the mention "all rights reserved" for scientific      */
/*           conference publication issues... Sorry.                                   */
/*                                                                                     */
/***************************************************************************************/

/***************************************************************************************/
/*                                                                                     */
/*
|=======================|
|Glue pour protocole RIP|
|=======================|


Setup (serveur) :
=================

# serveurRIP config.txt
(config.txt contient une liste de fichiers)

I.      - Trouver le fichier le plus gros de config.txt
       - Noter T sa taille et n le nombre de fichiers

II.     Créer un tableau avec chaque scalaire sur l_0 bits (rappel N=50 et l_0=16)
       - Si T*n > 1 Gigaoctet :
               - Créer un tableau 3D A[n][ceil(100 * 10^6 / (16*n*N*NLA)) * NLA ][N].
               - Inclure les fichiers avec le format suivant :
                       - Premiers 64 bits : Taille du fichier
                       - Reste : fichier + padding de zéros
               - Boucler jusqu'à arriver à la fin
       - Si T*n < 1 Gigaoctet :
               - Créer un tableau 4D A[ITER][n][ceil(100 * 10^6 / (16*n*N*NLA)) * NLA ][N]. Ou ITER est le nombre d'itérations pour finir avec tous les fichiers plus le padding.
               - Inclure les fichiers avec le format suivant :
                       - Premiers 64 bits : Taille du fichier
                       - Reste : fichier + padding de zéros
               - Boucler jusqu'à arriver à la fin

III.    Transférer un bloc de 100 MégaOCTETS vers la mémoire de la carte graphique.

IV.     Allouer C[ceil(100 * 10^6 / (16*n*N*NLA)) * NLA ][N] dans la carte graphique.

V.      Créer une structure DESC contenant la liste de fichiers et la taille ceil(T/(N*l_0)) * (N*l_0) d'un élément de la BDD


Réception de connexion (serveur) :
==================================


Écoute sur un port TCP

I.      Le client envoie un hello
II.     Le serveur répond en envoyant DESC
III.    Le client envoie un tableau B[n*N][2N]


Traitement (serveur) :
======================

I.      Mettre C à zéro
II.     Boucler sur
               -C+=Calcul(Ai,Bi)
               -remplacer Ai et Bi
III.    Renvoyer C et démarrer la prochaine itération


Connexion (client) :
====================

Connexion à un port TCP

I.      Encoyer un hello
II.     Proposer DESC à l'utilisateur et lui demander un choix
III.    Générer un requête RIP pour l'index demandé et l'envoyer au serveur
IV.     Récupérer la réponse RIP et la décoder en la stockant


Génération d'une requête RIP (client) :
=======================================

I.      Fixer p=2^47+5 et choisir q nombre aléatoire de 2l_0 bits (exactement)
II.     Obtenir la matrice mère M=[A|B] A,B matrices de N x N scalaires dans Z/pZ (i.e. rand mod 2^47 + 5)
       - B matrice aléatoire
       - A matrice aléatoire inversible
               - Choisir une matrice aléatoire
               - Tester si elle est inversible (fonction à fournir par Philippe)
               - Si non changer un scalaire et retester
               - Si oui la définir comme A
       - Générer n matrices M"_1 ... M"_n pour chacune faire :
               - Générer une matrice P_i N x N dans Z/pZ aléatoire inversible
               - Définir M"_i = P_i x M (Philippe fournit la fonction multiplication mod p)
       - Générer une matrice Delta N x N inversible dans Z/pZ
       - Générer n matrices M'_1 ... M'_n :
               - Générer des matrice D_i N x N dans {-1,1} aléatoires pour i de 1 à n
               - Modifier D_i0 en remplaçant ses termes diagonaux par +q (i0 étant l'index choisi par l'utilisateur)
               - Définir M'_i = M''_i + [0| D_i x Delta ]
       - Générer n matrices M_1 ... M_n :
               - Choisir aléatoirement une permutation P de [1,2N]
               - Calculer M_i = P (M'_i)
       - L'ensemble (ordonné) des M_i est la requête


Décodage de la réponse RIP (client):
====================================

I.      Noter V le vecteur de taille 2N réponse de la BDD
II.     Dé-permuter le vecteur V et noter V'= P^{-1} (V)
III.    Noter V'=[V_U|V_D] V_U et V_D étant les moitiés Undisturbed et Disturbed
IV.     Récupérer la deuxième moitié Undisturbed avec V_U2 = V_U x A^{-1} x B
V.      Obtenir le bruit du calcul RIP E = V_D - V_U2
VI.     Déamplifier le bruit E' = E x Delta^{-1}
VII.    Pour E' = [e'_1,...,e'_N] calculer e''_i = e'_i - epsilon, avec epsilon = e'_i % q si c'est plus petit que q/2 et epsilon = e'_i % q - q sinon
VIII.   Calculer le ième morceaux de l'élément m_i0 (l'element de la BDD à récupérer) par m_{i0,i}= e''_i * q^{-1} (fonction de calcul de l'inverse d'un scalaire et de la multiplication à fournir par Philippe)
*/
/***************************************************************************************/


#ifndef __COMMON_RIP__
#define __COMMON_RIP__

typedef long long int64;



// number of rows of the RIP query matrix
#define N           50

// maximum amount of data loadable at once in RIP data matrix (expressed in bytes, minimum 4k by data file) - 500Mo seems fine. Please note that server will occupy more than that (also has to store query and result, whose size depend on the file count)
#define MAX_HOST_MEMORY_FOR_DATA    (1024*1024*500)

// number of lines of the data matrix which will be computed per iteration
#define LINES_COMPUTED_AT_ONCE      8000
//#define LINES_COMPUTED_AT_ONCE      10



// port to which the server is bound
#define SERVER_PORT 7660



#endif // __COMMON_RIP__






