/***************************************************************************************/
/*                                                                                     */
/*           RIP protocol : client program                                             */
/*                                                                                     */
/***************************************************************************************/
/***************************************************************************************/
/*                                                                                     */
/*           Pierre Rousseau and Carlos Aguilar Melchor.                               */
/*           Limoges University, France.                                               */
/*           All rights reserved.                                                      */
/*           Note: once our work is finished all this work will be under a GPL licence.*/
/*           Until then, we keep the mention "all rights reserved" for scientific      */
/*           conference publication issues... Sorry.                                   */
/*                                                                                     */
/***************************************************************************************/



#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

#include<time.h>

#include <fstream>
#include <string> 
#include <iostream>
#include <vector>
#include <cmath>
using namespace std;


#include "CommonRIP.h"



#define SERVER_IP   "127.0.0.1"

#include <NTL/mat_ZZ_p.h>



NTL_CLIENT





/////////////////////////////////////////////////
/////////////////////////////////////////////////
// int64 functions
/////////////////////////////////////////////////
/////////////////////////////////////////////////
ZZ_p int64ToZZ_p(int64 a)
{
    ZZ tmp;
    tmp = ZZFromBytes(reinterpret_cast<unsigned char *>(&a), 8);
    ZZ_p ret;
    conv (ret, tmp);
    return ret;
}

int64 ZZ_pToInt64(ZZ_p a)
{
    unsigned char *glop = (unsigned char *)malloc(8);
    ZZ tmp = rep(a);
    BytesFromZZ(glop, tmp, 8);
    int64 ret = *reinterpret_cast<int64*>(glop);
    free(glop);
    return ret;
}

ZZ int64ToZZ(int64 a)
{
    ZZ tmp;
    tmp = ZZFromBytes(reinterpret_cast<unsigned char *>(&a), 8);
    return tmp;
}

int64 ZZToInt64(ZZ a)
{
    unsigned char *glop = (unsigned char *)malloc(8);
    BytesFromZZ(glop, a, 8);
    int64 ret = *reinterpret_cast<int64*>(glop);
    free(glop);
    return ret;
}


int64 random_to(int64 max)
{
    int64 a = (int64)( ( (int64)rand() << 32 ) + ( (int64)rand() ) ) % max;
    return a;
}

int64 random_n_bits(int num_bits)
{
    int64 a = ( ((int64)1<<num_bits) + random_to((int64)1<<num_bits) );
    int64 b = ((int64)1<<(num_bits+1));
    int64 c = a%b;
    
    return c;
}


int64 multMod(int64 a, int64 b, int64 module)
{
    unsigned char res_char[4];
    ZZ_p a_zz, b_zz, r_zz;
    ZZ n_zz;
    
    a_zz = int64ToZZ_p(a);
    b_zz = int64ToZZ_p(b);
    n_zz = int64ToZZ(module);
    
    ZZ_pBak bak;  
    bak.save();  // save current modulus in bak

    ZZ_p::init(n_zz);  // set the current modulus to module
    mul(r_zz, a_zz, b_zz);
    
    return ZZ_pToInt64(r_zz);
}






vector<int> reversePermutation(vector<int> P)
{
    vector<int> P2;

    P2 = P; // just to make sure there's the exact space needed.
    
    int i = 0;
    
    for (vector<int>::iterator it = P.begin() ; it < P.end() ; it++ , i++)
    {
        P2[*it] = i;
    }
    return P2;
}


/////////////////////////////////////////////////
/////////////////////////////////////////////////
// Matrix_int64 class
/////////////////////////////////////////////////
/////////////////////////////////////////////////
class Matrix_int64
{
    public :
        int num_rows;
        int num_cols;
        //internal representation :
        int64 *M;
        mat_ZZ_p *M_zz;
    
        Matrix_int64(int _num_rows, int _num_cols, int64 *_M = NULL)
        {
            num_rows = _num_rows;
            num_cols = _num_cols;
            if (_M == NULL)
            {
                M = (int64 *)malloc(sizeof(int64) * num_rows * num_cols);
                initialize();
            }
            else
            {
                M = _M;
            }
            M_zz = NULL;
        }
        
        
        ~Matrix_int64()
        {
            free(M);
            if(M_zz != NULL)
            {
                delete(M_zz);
                M_zz = NULL;
            }
        }
        
        
        void setValue(int64 val, int row, int col)
        {
            if (row < num_rows && col < num_cols && row >= 0 && col >= 0)
            {
                M[row*num_cols + col] = val;
                
                if (M_zz != NULL)
                {
                    (*M_zz)[row][col] = int64ToZZ_p(val);
                }
            }
        }
        
        void addValue(int64 val, int row, int col)
        {
            if (row < num_rows && col < num_cols && row >= 0 && col >= 0)
            {
                M[row*num_cols + col] += val;
                
                if (M_zz != NULL)
                {
                    (*M_zz)[row][col] += int64ToZZ_p(val);
                }
            }
        }
        
        int64 getValue(int row, int col)
        {
            if (row < num_rows && col < num_cols && row >= 0 && col >= 0)
            {
                return M[row*num_cols + col];
            }
            else
            {
                // matrix cell doesn't exist !
                return 0;
            }
        }
        
        
        void initialize()
        {   
            bzero(M, sizeof(int64) * num_rows * num_cols);
        }
        
        
        void fillRandomly(int64 max_val)
        {
            int64 *tmp = M;
            
            bzero(tmp, sizeof(int64) * num_rows * num_cols);
            
            for (int r = 0 ; r < num_rows ; r++)
            {
                for (int c = 0 ; c < num_cols ; c++)
                {
                    *tmp++ = random_to(max_val);
                }
            }
            
            // invalidate M_zz if it exists
            if (M_zz != NULL)
            {
                delete (M_zz);
                M_zz = NULL;
            }
        }
        
        
        void alterRandom (int64 max_val)
        {
            int rand_col = random_to(num_cols);
            int rand_row = random_to(num_rows);
            
            int64 new_val = random_to(max_val);
            M[num_cols*rand_row + rand_col] = new_val;
            
            if (M_zz != NULL)
            {
                (*M_zz)[rand_row][rand_col] = int64ToZZ_p(new_val);
            }
        }
        
        
        void toMatZZ_p()
        {
            if (M_zz != NULL)
            {
                delete (M_zz);
            }
            
            M_zz = new mat_ZZ_p;
            M_zz->SetDims(num_rows, num_cols);
            
            int64 *tmpMat = M;
            
            for (int r = 0 ; r < num_rows ; r++)
            {
                for (int c = 0 ; c < num_cols ; c++)
                {
                    int64 glop = *tmpMat++;
                    (*M_zz)[r][c] = int64ToZZ_p(glop);
                }
            }
        }
        
        
        void toMatInt64()
        {
            if (M_zz == NULL)
            {
                return;
            }
            
            for (int r = 0 ; r < num_rows ; r++)
            {
                for (int c = 0 ; c < num_cols ; c++)
                {
                    int64 tmp = ZZ_pToInt64( (*M_zz)[r][c]);
                    M[r*num_cols + c] = tmp;
                }
            }
        }
        
        
        void setMatZZ_p(const mat_ZZ_p & A)
        {
            // if M_zz already exist, we MUST delete it !
            if (M_zz != NULL)
            {
                delete(M_zz);
            }
            
            M_zz = new mat_ZZ_p(A);
            
            // now copy this to the int64 * matrix;
            toMatInt64();
        }
        
        
        
        
        bool is_invertible ()
        {
            if (M_zz == NULL)
            {
                toMatZZ_p();
            }
            
            // now compute the determinant to know if the matrix is invertible
            if (determinant(*M_zz) != 0)
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        
        
        /////////////////////////////////////////////////////////////////////
        // Static methods :
        /////////////////////////////////////////////////////////////////////
        
        
        
        
        static Matrix_int64* createRandomMatrix(int num_rows, int num_columns, int64 max_val)
        {
            Matrix_int64 *mat = new Matrix_int64(num_rows, num_columns);
            mat->fillRandomly(max_val);
            return mat;
        }
        
        
        
        static Matrix_int64* matrixConcat(Matrix_int64 *Left, Matrix_int64 *Right)
        {
            if (Left->num_rows != Right->num_rows)
            {
                cout << "Matrix concatenation impossible !" << endl;
                return NULL;
            }
            else
            {
                Matrix_int64 *ret = new Matrix_int64(Left->num_rows, Left->num_cols + Right->num_cols);
                
                for (int row = 0 ; row<Left->num_rows ; row++)
                {
                    for (int col = 0 ; col<Left->num_cols ; col++)
                    {
                        ret->setValue( (Left->getValue(row, col) ), row, col);
                    }
                    for (int col = 0 ; col<Right->num_cols ; col++)
                    {
                        ret->setValue( (Right->getValue(row, col) ), row, col+Left->num_cols);
                    }
                }
                
                return ret;
            }
        }
        
        
        // start and stop columns are included in the range.
        static Matrix_int64* extractColumns(Matrix_int64 *From, int start_col, int stop_col)
        {
            if (start_col>stop_col || start_col<0 || stop_col >= From->num_cols)
            {
                cout << "Column extraction impossible !" << endl;
                return NULL;
            }
            else
            {
#ifdef DEBUG
                cout << "col extraction : " << From->num_rows << " " <<  stop_col << " " << start_col << " " << stop_col - start_col << endl;
#endif
                Matrix_int64 *ret = new Matrix_int64(From->num_rows, stop_col - start_col + 1);
                
                for (int row = 0 ; row<ret->num_rows ; row++)
                {
                    for (int col = 0 ; col<=stop_col-start_col ; col++)
                    {
                        ret->setValue( (From->getValue(row, col+start_col) ), row, col);
                    }
                }
                
                return ret;
            }
        }
        
        
        static Matrix_int64* matrixMul(Matrix_int64 *A, Matrix_int64 *B)
        {
            if (A->num_cols != B->num_rows)
            {
                cout << "Matrix multiplication impossible !" << endl;
                return NULL;
            }
            else
            {
#ifdef DEBUG
                cout << "mult : A : " << A->num_rows << " x " << A->num_cols << " - B : " << B->num_rows << " x " << B->num_cols << endl;
#endif
                Matrix_int64 *ret = new Matrix_int64(A->num_rows, B->num_cols);
                
                for (int row = 0 ; row<A->num_rows ; row++)
                {
                    for (int col = 0 ; col<B->num_cols ; col++)
                    {
                        for (int elem = 0 ; elem < A->num_cols ; elem++)
                        {
                            ret->addValue( (A->getValue(row, elem) * B->getValue(elem, col) ), row, col);
                        }
                    }
                }
                
                return ret;
            }
        }
        
        
        static Matrix_int64* matrixMulMod(Matrix_int64 *A, Matrix_int64 *B, int64 module)
        {
            if (A->num_cols != B->num_rows)
            {
                cout << "Modular matrix multiplication impossible !" << endl << "A : " << A->num_rows << " x " << A->num_cols << " - B : " << B->num_rows << " x " << B->num_cols ;
                return NULL;
            }
            else
            {
#ifdef DEBUG
                cout << "multmod : A : " << A->num_rows << " x " << A->num_cols << " - B : " << B->num_rows << " x " << B->num_cols << endl;
#endif
                ZZ_pBak bak;  
                bak.save();  // save current modulus in bak
                ZZ_p::init(int64ToZZ(module));  // set the current modulus to module
                
                Matrix_int64 *ret = new Matrix_int64(A->num_rows, B->num_cols);
                
                A->toMatZZ_p();
                B->toMatZZ_p();
                ret->setMatZZ_p(*(A->M_zz) * *(B->M_zz));
                
                return ret;
            }
        }
        
        
        
        static Matrix_int64* matrixAdd(Matrix_int64 *A, Matrix_int64 *B)
        {
            if (A->num_rows != B->num_rows || A->num_cols != B->num_cols)
            {
                cout << "Matrix addition impossible !" << endl;
                return NULL;
            }
            else
            {
#ifdef DEBUG
                cout << "add : A : " << A->num_rows << " x " << A->num_cols << " - B : " << B->num_rows << " x " << B->num_cols << endl;
#endif
                Matrix_int64 *ret = new Matrix_int64(A->num_rows, A->num_cols);
                
                for (int row = 0 ; row<ret->num_rows ; row++)
                {
                    for (int col = 0 ; col<ret->num_cols ; col++)
                    {
                        ret->setValue( (A->getValue(row, col) + B->getValue(row, col) ), row, col);
                    }
                }
                
                return ret;
            }
        }
        
        
        static Matrix_int64* matrixAddMod(Matrix_int64 *A, Matrix_int64 *B, int64 module)
        {
            if (A->num_rows != B->num_rows || A->num_cols != B->num_cols)
            {
                cout << "Matrix addition impossible !" << endl;
                return NULL;
            }
            else
            {
#ifdef DEBUG
                cout << "add : A : " << A->num_rows << " x " << A->num_cols << " - B : " << B->num_rows << " x " << B->num_cols << endl;
#endif
                ZZ_pBak bak;  
                bak.save();  // save current modulus in bak
                ZZ_p::init(int64ToZZ(module));  // set the current modulus to module
                
                Matrix_int64 *ret = new Matrix_int64(A->num_rows, B->num_cols);
                
                A->toMatZZ_p();
                B->toMatZZ_p();
                ret->setMatZZ_p(*(A->M_zz) + *(B->M_zz));
                
                return ret;
            }
        }
        
        
        
        static Matrix_int64* matrixSubtract(Matrix_int64 *A, Matrix_int64 *B)
        {
            if (A->num_rows != B->num_rows || A->num_cols != B->num_cols)
            {
                cout << "Matrix subtraction impossible !" << endl;
                return NULL;
            }
            else
            {
#ifdef DEBUG
                cout << "sub : A : " << A->num_rows << " x " << A->num_cols << " - B : " << B->num_rows << " x " << B->num_cols << endl;
#endif
                Matrix_int64 *ret = new Matrix_int64(A->num_rows, B->num_cols);
                
                for (int row = 0 ; row<ret->num_rows ; row++)
                {
                    for (int col = 0 ; col<ret->num_cols ; col++)
                    {
                        ret->setValue( A->getValue(row, col) - B->getValue(row, col), row, col);
                    }
                }
                
                return ret;
            }
        }
        
        
        static Matrix_int64* matrixSubtractMod(Matrix_int64 *A, Matrix_int64 *B, int64 module)
        {
            if (A->num_rows != B->num_rows || A->num_cols != B->num_cols)
            {
                cout << "Matrix subtraction impossible !" << endl;
                return NULL;
            }
            else
            {
#ifdef DEBUG
                cout << "sub : A : " << A->num_rows << " x " << A->num_cols << " - B : " << B->num_rows << " x " << B->num_cols << endl;
#endif
                ZZ_pBak bak;  
                bak.save();  // save current modulus in bak
                ZZ_p::init(int64ToZZ(module));  // set the current modulus to module
                
                Matrix_int64 *ret = new Matrix_int64(A->num_rows, B->num_cols);
                
                A->toMatZZ_p();
                B->toMatZZ_p();
                ret->setMatZZ_p(*(A->M_zz) - *(B->M_zz));
                
                return ret;
            }
        }
        
        
        
        static Matrix_int64 *permutation(Matrix_int64 *originalMatrix, const vector<int> & P)
        {
            Matrix_int64 *ret = new Matrix_int64(originalMatrix->num_rows, originalMatrix->num_cols);
            
            for (int row = 0 ; row<ret->num_rows ; row++)
            {
                for (int col = 0 ; col<ret->num_cols ; col++)
                {
                    ret->setValue( originalMatrix->getValue(row, P[col]), row, col);
                }
            }
            
            return ret;
        }
        
        
        static Matrix_int64 *invertMod(Matrix_int64 *originalMatrix, int64 module)
        {
            if (!originalMatrix->is_invertible())
            {
                cout << "Matrix inversion impossible !" << endl;
                return NULL;
            }
            else
            {
#ifdef DEBUG            
                cout << "inv : A : " << originalMatrix->num_rows << " x " << originalMatrix->num_cols  << endl;
#endif
                ZZ_pBak bak;  
                bak.save();  // save current modulus in bak
                ZZ_p::init(int64ToZZ(module));  // set the current modulus to module
                
                Matrix_int64 *ret = new Matrix_int64(originalMatrix->num_rows, originalMatrix->num_cols);
                
                ret->setMatZZ_p( inv(* (originalMatrix->M_zz) ) );
                
                return ret;
            }
        }
        
};














//////////////////////////////////////////////////////////////
/// This class creates the 'B' matrix, database query for the server
//////////////////////////////////////////////////////////////
class RIPQuery
{
    public :
        int64 p; // prime number, used in computation process
        int64 q; // random number, 0<=q<=p
        int chosen_element; // selected index
        int num_elems;
        int l0; // number of bits
        
        // the matrices :
        Matrix_int64 *A, *B;
        Matrix_int64 *delta;
        vector<Matrix_int64 *> P_i_list;
        vector<Matrix_int64 *> M_second_i_list;
        vector<Matrix_int64 *> D_i_list;
        vector<Matrix_int64 *> M_prime_i_list;
        vector<Matrix_int64 *> M_i_list; //this is the query !
        vector<int> P; // permutation list
        
        
        RIPQuery(int _chosen_element, int _num_elems)
        {        
            cout << "Creating database query" << endl;
            
            chosen_element = _chosen_element;
            num_elems = _num_elems;
            l0 = 16;
            
            p = ((int64)1<<47) + 5;
            ZZ_p::init(int64ToZZ(p));
            
            //q = random_n_bits(2*l0);
            q = random_n_bits(30);
            
            //cout << "q : " << q << endl;
            
            //////////////////////////////////////////////////////////////////
            //      Mother Matrix, M
            //////////////////////////////////////////////////////////////////
            
            // prepare 'mother' matrix, M, composed of A and B.
            A = Matrix_int64::createRandomMatrix(N, N, p);
            while (! A->is_invertible() )
            {
                A->alterRandom(p);
            }
            
            B = Matrix_int64::createRandomMatrix(N, N, p);
            
            // creation of the 'mother' matrix, M. This is [A|B] (matrix concatenation)
            Matrix_int64 *M = Matrix_int64::matrixConcat(A, B);
            
            
            //////////////////////////////////////////////////////////////////
            //      M_second_i matrices
            //////////////////////////////////////////////////////////////////
            
            // generation of the M''_i matrices
            for (int i = 0 ; i < num_elems ; i++)
            {
                Matrix_int64 *P_i = Matrix_int64::createRandomMatrix(N, N, p);
                
                while (! P_i->is_invertible() )
                {
                    P_i->alterRandom(p);
                }
                
                // save this P_i
                P_i_list.push_back(P_i);
                
                // now compute M''_i : M''_i = P_i * M (multiplication mod p)
                Matrix_int64 *M_second_i = Matrix_int64::matrixMulMod(P_i, M, p);
                M_second_i_list.push_back(M_second_i);
            }
            
            // generation of the 'delta' matrix
            delta = Matrix_int64::createRandomMatrix(N, N, p);
            
            while (! delta->is_invertible() )
            {
                delta->alterRandom(p);
            }
            
            //////////////////////////////////////////////////////////////////
            //      M_prime_i matrices
            //////////////////////////////////////////////////////////////////
            
            // generation of the M'_i matrices
            for (int i = 0 ; i < num_elems ; i++)
            {
                Matrix_int64 *D_i = new Matrix_int64(N, N);
                for (int row = 0 ; row<N ; row++)
                {
                    for (int col = 0 ; col<N ; col++)
                    {
                        D_i->setValue( ((rand()%2)==0)?1:p-1 , row, col);
                    }
                }
                
                // the D_i matrix associated to the file the user chose, deserves a special treatment : set its diagonal to q
                if (i == chosen_element)
                {
                    for (int diag = 0 ; diag < N ; diag++)
                    {
                        D_i->setValue(q, diag, diag);
                    }
                }
                
                // save this D_i
                D_i_list.push_back(D_i);
                
                
                // now M'_i is obtained by : M'_i = M''-i + [ 0 | D_i * delta] ;
                
                // multiply D_i by delta
                Matrix_int64 *D_i_delta = Matrix_int64::matrixMulMod(D_i, delta, p);
                
                Matrix_int64 *zero_matrix = new Matrix_int64(N, N);
                
                Matrix_int64 *zero_D_i_delta = Matrix_int64::matrixConcat(zero_matrix, D_i_delta);
                
                Matrix_int64 *M_prime_i = Matrix_int64::matrixAddMod(M_second_i_list[i], zero_D_i_delta, p);
                
                delete(D_i_delta);
                delete(zero_matrix);
                delete(zero_D_i_delta);
                
                M_prime_i_list.push_back(M_prime_i);
                
            }
            
            
            //////////////////////////////////////////////////////////////////
            //      M_i matrices
            //////////////////////////////////////////////////////////////////
            
            // determine a permutation order, to move the columns
            // enter the positions in order
            for (int i = 0 ; i < 2*N ; i++)
            {
                P.push_back(i);
            }
            // now scramble it !
            for (int i = 0 ; i < 200 ; i++) // do 200 random permutations
            {
                int ind_1 = rand() % (2*N);
                int ind_2 = rand() % (2*N);
                int tmp = P[ind_1];
                P[ind_1] = P[ind_2];
                P[ind_2] = tmp;
            }
            
            // generation of the M_i matrices : M_i = P(M'_i)
            for (int i = 0 ; i < num_elems ; i++)
            {
                Matrix_int64 *M_i = Matrix_int64::permutation(M_prime_i_list[i], P);
                M_i_list.push_back(M_i);
            }
            
            delete(M);
            cout << "Query created !" << endl;
        }
        
        
        
        ~RIPQuery()
        {
            delete(A);
            delete(B);
            delete(delta);
            
            for (vector<Matrix_int64 *>::iterator it = P_i_list.begin() ; it < P_i_list.end() ; it++)
            {
                delete(*it);
            }
            P_i_list.clear();
            
            for (vector<Matrix_int64 *>::iterator it = M_second_i_list.begin() ; it < M_second_i_list.end() ; it++)
            {
                delete(*it);
            }
            M_second_i_list.clear();
            
            for (vector<Matrix_int64 *>::iterator it = D_i_list.begin() ; it < D_i_list.end() ; it++)
            {
                delete(*it);
            }
            D_i_list.clear();
            
            for (vector<Matrix_int64 *>::iterator it = M_prime_i_list.begin() ; it < M_prime_i_list.end() ; it++)
            {
                delete(*it);
            }
            M_prime_i_list.clear();
            
            for (vector<Matrix_int64 *>::iterator it = M_i_list.begin() ; it < M_i_list.end() ; it++)
            {
                delete(*it);
            }
            M_i_list.clear();
        }
};









//////////////////////////////////////////////////////////////
/// This class processes the file list transmitted
/// from the server
//////////////////////////////////////////////////////////////
class DESC
{
    public :
        vector<string> fileList;
        int fileCount;
        int resultRows, resultCols, resultCount;
        
        DESC (char * receivedBuffer)
        {
            int alreadyRead = 0;
            int match = 0;
            char fileCountSize[10];
            
            if (sscanf(receivedBuffer, "%d %d %d %d", &fileCount, &resultRows, &resultCols, &resultCount) != 0)
            {
                sscanf(receivedBuffer, "%[^\n]", fileCountSize);
                alreadyRead += strlen(fileCountSize) + 1;
                
                for (int i = 0 ; i < fileCount ; i++)
                {
                    char fileName[1024];
                    match = sscanf(receivedBuffer+alreadyRead, "%[^\n]", fileName);
                    if (match != 0)
                    {
                        alreadyRead += strlen(fileName) + 1; // +1 : read the EOL
                        fileList.push_back(fileName);
                    }
                }
            }
            else
            {
                cout << "Impossible to get file list from the server. Retrying in 1 second." << endl;
                fileCount = -1;
            }
        }
        
        
        string getFileName(int index)
        {
            return fileList[index];
        }
        
        
        void proposeMenu()
        {
            cout << "##############################################" << endl;
            cout << "#" << endl;
            cout << "# Connection established" << endl;
            cout << "#" << endl;
            cout << "##############################################" << endl;
            cout << "#" << endl;
            cout << "# File List :" << endl;
            cout << "# °°°°°°°°°°°" << endl;
            cout << "#" << endl;
            
            for (int i = 0 ; i < fileCount ; i++)
            {
                cout << "# " << i+1 << ") " << getFileName(i) << endl;
            }
            
            cout << "#" << endl;
            cout << "##############################################" << endl;
            cout << "#" << endl;
            cout << "# Which file do you want ?" << endl;
        }
        
        // this function checks whether the provided file number exists in the DESC sended by the server
        bool file_exists (int ID)
        {
            if (ID >= 0 && ID < fileCount)
            {
                return true;
            }
            else
            {
                return false;
            }
        }
};






class RIPResult
{
    public :
    
        //Matrix_int64* Vi; // temporary result from the server
        vector<Matrix_int64*> V; // temporary result from the server
        Matrix_int64* Mi; // decoded result, to be written to file
        // precomputed inverted matrix, dependant on the query :
        Matrix_int64 *A_inv;
        Matrix_int64 *delta_inv;
        
        RIPQuery *query;
        DESC *catalogue;
        vector<int> reverse_perm;
        
        int64 fileSize;
        string fileName;
        
        // constructor
        RIPResult(RIPQuery *_query, DESC *_catalogue)
        {
            query = _query;
            catalogue = _catalogue;
            
            // get reverse permutation
            reverse_perm = reversePermutation(query->P);
            
            // we'll wonder about the name later
            fileName = string("received_files/") + catalogue->getFileName(query->chosen_element);
            
            // check if file exists
            fstream f(fileName.c_str());
            if (f)
            {
                f.close();
                
                cout << endl;
                cout << endl;
                cout << "###########################" << endl;
                cout << "##" << endl;
                cout << "## FILE ALREADY EXISTS !!!" << endl;
                cout << "##" << endl;
                cout << "## SUPPRESSING IT !" << endl;
                cout << "##" << endl;
                cout << "###########################" << endl;
                cout << endl;
                cout << endl;
                
                remove(fileName.c_str());
            }
            
            fileSize = 0;
            // no result currently received
            //Vi = NULL;
            Mi = NULL;
            
            // some pre-computation :
            A_inv = Matrix_int64::invertMod(query->A, query->p);
            delta_inv = Matrix_int64::invertMod(query->delta, query->p);
        }
        
        
        ~RIPResult()
        {
            // Vi and Mi should already have been deallocated
            if (Mi != NULL)
            {
                delete(Mi);
                Mi = NULL;
            }
            /*
            if (Vi != NULL)
            {
                delete(Vi);
                Vi = NULL;
            }
            */
            for (vector<Matrix_int64 *>::iterator it = V.begin() ; it < V.end() ; it++)
            {
                if (*it != NULL)
                {
                    delete (*it);
                    *it = NULL;
                }
            }
            V.clear();
            
            delete(A_inv);
            delete(delta_inv);
        }
        
        
        
        
        void writeFile()
        {
            std::fstream f(fileName.c_str(), std::ios_base::out | std::ios_base::binary | std::ios_base::app);
            long pos = 0;
            long loadableSize = sizeof(int64) * Mi->num_rows * Mi->num_cols;
            int64 *tmp = Mi->M;
            
            if (f)
            {
                f.seekp( 0, std::ios_base::end );
                long alreadyWritten = f.tellp() ;
                
                if (alreadyWritten == 0)
                {
                    // we are beginning to write the file.
                    // first : extract its size !
                    fileSize = 0;
                    int64 i = 0;
                    
                    i = (unsigned short)(*tmp++);
                    fileSize += i<<48;
                    i = (unsigned short)(*tmp++);
                    fileSize += i<<32;
                    i = (unsigned short)(*tmp++);
                    fileSize += i<<16;
                    i = (unsigned short)(*tmp++);
                    fileSize += i;
                    
                    //printf("PQ : %llX\n" ,  query->p / query ->q);
                    
                    //cout << "size decoded : " << fileSize << endl;
                    
                    //some data has already been extracted from the matrix
                    pos = 32;   // we already extracted 4 x int64
                }
                
                while (alreadyWritten < fileSize && pos < loadableSize)
                {
                    f.write(reinterpret_cast<const char*>(tmp++), sizeof(unsigned short));
                    pos += 8; // we just read an int64 -> 8 bytes
                    alreadyWritten += 2; // only the last 2 bytes contain actual written data.
                    
                    //using shorts might be a problem in some cases (text files with odd character count for example..)
                    if (alreadyWritten == fileSize -1)
                    {
                        f.write(reinterpret_cast<const char*>(tmp++), sizeof(char));
                        alreadyWritten++; // only wrote one byte
                        pos+=8; // read an int64
                    }
                }
                f.close();
                cout << "Already written : " << alreadyWritten/(1024.0*1024.0) << " Mo / " << fileSize/(1024.0*1024.0) << " Mo (" << alreadyWritten * 100.0 / float(fileSize) << "%)." << endl;
                
                // this decoded result is written, we no longer need to store it.
                delete(Mi);
                Mi = NULL;
            }
            else
            {
                cout << "Couldn't open output file for writing." << endl;
                exit(1);
            }
        }
        
        
        
        void addResult(Matrix_int64 *res)
        {
            //Vi = res;
            V.push_back(res);
        }
        
        
        void decode(int num_elem)
        {
            //cout << "Starting to decode the partial message." << endl;
            
            // apply reverse permutation to the result
            //cout << "Permutation " << endl;
            
            //Matrix_int64 *V_received = V.back();
            //Matrix_int64 *V_received = Vi;
            Matrix_int64 *V_received = V[num_elem];
            
            
            
            Matrix_int64 *V_prime = Matrix_int64::permutation(V_received, reverse_perm);  // use last partial result, i.e. the one which we most lately received.
            
            // now V_prime can be written as : V'=[V_U|V_D]   -- each part being undisturbed and disturbed part
            // extract V_U and V_D
            //cout << "Sub-matrices extraction" << endl;
            Matrix_int64 *V_U = Matrix_int64::extractColumns(V_prime, 0, (V_prime->num_cols / 2) - 1);
            Matrix_int64 *V_D = Matrix_int64::extractColumns(V_prime, V_prime->num_cols / 2, V_prime->num_cols - 1);
            
            // compute V_U2 = V_U x A^{-1} x B
            //cout << "V_U2 = V_U x A^{-1} x B" << endl;
            
            // A_inv can be pre-computed, it doesn"t vary with the result.
            //Matrix_int64 *A_inv = Matrix_int64::invertMod(query->A, query->p);
            
            Matrix_int64 *V_U_A_inv = Matrix_int64::matrixMulMod(V_U, A_inv, query->p);
            Matrix_int64 *V_U2 = Matrix_int64::matrixMulMod(V_U_A_inv, query->B, query->p);
            
            // extract noise : E = V_D - V_U2
            //cout << "V_D - V_U2" << endl;
            Matrix_int64 *E = Matrix_int64::matrixSubtractMod(V_D, V_U2, query->p);
            
            // noise is amplificated. de-amplificate it now : E' = E x Delta^{-1}
            //cout << "Noise deamplification" << endl;
            
            // delta_inv can be pre-computed, it doesn"t vary with the result.
            //Matrix_int64 *delta_inv = Matrix_int64::invertMod(query->delta, query->p);
            Matrix_int64 *E_prime = Matrix_int64::matrixMulMod(E, delta_inv, query->p);
            
            // FROM THIS POINT, OPERATIONS ARE NO LONGER MODULAR
            
            Matrix_int64 * m_i0 = new Matrix_int64(E_prime->num_rows, 0); // this just defines the matrix size, no data inserted
            
            // for each column e_prime_i of E_prime, compute e''_i = e'_i - epsilon, with epsilon = e'_i % q
            for (int i = 0 ; i < E_prime->num_cols ; i++)
            {
                Matrix_int64 * e_prime_i = Matrix_int64::extractColumns(E_prime, i, i);
                Matrix_int64 * epsilon = new Matrix_int64(e_prime_i->num_rows, 1);
                for (int r=0 ; r < epsilon->num_rows ; r++)
                {
                    int64 eps_j = (e_prime_i->getValue(r, 0)) % query->q;
                    if (eps_j > query->q/2)
                    {
                        eps_j -= query->q;
                    }
                    epsilon->setValue(eps_j, r, 0);
                }
                
                Matrix_int64 * e_second_i = Matrix_int64::matrixSubtract(e_prime_i, epsilon);
                
                // compute m_{i0,i}= e''_i * q^{-1}
                Matrix_int64 * m_i0_i = new Matrix_int64(e_prime_i->num_rows, 1);
                for (int r=0 ; r < m_i0_i->num_rows ; r++)
                {
                    int64 tmp = (e_second_i->getValue(r, 0) % query->p) / query->q;
                    int64 tmp2 = tmp % (query->p / query->q);
                    m_i0_i->setValue(tmp2 , r, 0);
                }
                
                // concatenate m_i0 :
                Matrix_int64 * tmpConcat = m_i0;
                m_i0 = Matrix_int64::matrixConcat(tmpConcat, m_i0_i); // tadaaaaa !
                
                delete(e_second_i);
                delete(tmpConcat);
                delete(m_i0_i);
                
                delete(e_prime_i);
                delete(epsilon);
            }
            
            Mi = m_i0;
            
            //cout << "Partial message decoding done." << endl;
            
            // now some cleaning :
            delete(V_prime);
            delete(V_U);
            delete(V_D);
            delete(V_U_A_inv);
            delete(V_U2);
            delete(E);
            delete(E_prime);
            
            // we can erase the received result, as it is now decoded
            /*
            delete(Vi);
            Vi = NULL;
            */
        }
        
        
        
        
        
};




class RIPClient
{
    public :

        int s, domain, type, protocol;
        sockaddr_in server_info;
        struct sockaddr *cliaddr;
        char buffer[65535];
        bool retry;
        DESC *catalogue;
        RIPQuery *query;
        RIPResult *result;
        
        
        // constructor
        RIPClient()
        {
            do 
            {
                startNewConnection();
                
                // receive file list
                int charactersRead=recv(s, &buffer, sizeof(buffer),0);
                
                catalogue = new DESC(buffer);
                
                if (catalogue->fileCount < 0)
                {
                    closeConnection();
                    retry = true;
                    sleep (1);
                }
                else
                {
                    retry = false;
                }
            }
            while (retry);
            
            int chosenFileID;
            
            catalogue->proposeMenu();
            do
            {
                cin >> chosenFileID;
                chosenFileID--; // files are indexed from 0 to n-1
                if (catalogue->file_exists(chosenFileID))
                {
                    retry = false;
                }
                else
                {
                    cout << "# This file doesn't exist !!! " << endl;
                    cout << "# Which file do you want ?" << endl;
                    retry = true;
                }
            }
            while (retry);
            
            cout << "You asked to receive file number " << chosenFileID+1 << ", \"" << catalogue->getFileName(chosenFileID) << "\"" << endl;
            
            query = new RIPQuery(chosenFileID, catalogue->fileCount);
            
            // let's upload the query !
            uploadQuery();
            
            result = new RIPResult(query, catalogue);
            /*
            for (int i = 0 ; i < catalogue->resultCount ; i++)
            {
                cout << "Waiting for result (" << i+1 << "/" << catalogue->resultCount << ")" << endl;
                
                getResult();
                
                result->decode();
                
                result->writeFile();
            }
            */
            
            for (int i = 0 ; i < catalogue->resultCount ; i++)
            {
                cout << "Waiting for result (" << i+1 << "/" << catalogue->resultCount << ")" << endl;
                
                getResult();
            }
            cout << "Received everything, starting to decode." << endl;
            for (int i = 0 ; i < catalogue->resultCount ; i++)
            {
                result->decode(i);
                
                result->writeFile();
            }
            
            cout << endl << "File \"" << result->fileName << "\" successfully received." << endl << endl;
        }
        
        
        
        ~RIPClient()
        {
            delete(catalogue);
            delete(query);
            delete(result);
        }
        
        
        
        void startNewConnection()
        {
            int error= -1;
            
            cout << "Initiating client." << endl;
            
            domain = AF_INET;
            type = SOCK_STREAM;
            protocol = 0;
            
            // socket creation
            s = socket (domain, type, protocol);
            
            // socket binding
            server_info.sin_family = AF_INET;
            server_info.sin_port = htons(SERVER_PORT); // the port we'll connect to
            int32_t i32Res = inet_pton(AF_INET, SERVER_IP, (void *) &server_info.sin_addr);
         
            if(0 > i32Res)
            {
                printf("error: first parameter is not a valid address family");
                exit(-1);
            }
            else if(0 == i32Res)
            {
                printf("char string (second parameter does not contain valid ipaddress");
                exit(-1);
            }
        
        
            error = connect(s,(struct sockaddr*) &server_info, sizeof(server_info));
            if (error==-1)
            {
                cout << "Connection failed." << endl;
                exit(-1);
            }
            
            // make sure the buffer is all zeros
            bzero(buffer, sizeof(buffer));
            strcpy(buffer,"Hello"); // prepare buffer for transmission to server. First message : "Hello"
            int charactersWritten=send(s,buffer,strlen(buffer),0);
            if (charactersWritten != strlen(buffer))
            {
                //cout << "Couldn't send Hello to server." << endl;
                exit (1);
            }
            else
            {
                //cout << "Hello sent." << endl;
            }
        }
        
        
        void closeConnection()
        {
            shutdown(s, 2);
            close(s);
        }
        
        
        
        void uploadQuery()
        {
            int bufferSize = N * 2*N * sizeof(int64);
            char *buffer = (char *)malloc(bufferSize);
            
            //cout << "Uploading query." << endl;
            
            // first : upload the modulus to be used.
            char moduleBuffer[8];
            memcpy(moduleBuffer, &(query->p), 8);
            send(s, moduleBuffer, 8, 0);
            
            // second : upload the data
            for (int i = 0 ; i < query->num_elems ; i++)
            {
                bzero(buffer, bufferSize);
                memcpy(buffer, query->M_i_list[i]->M, bufferSize);
                
                // we can't upload all at once
                // let's split it in sections of 1k
                int uploaded = 0;
                int uploadSize = 1024;
                while (uploaded != bufferSize)
                {
                    uploadSize = (uploaded + uploadSize) < bufferSize ? uploadSize : (bufferSize - uploaded);
                    int charactersWritten=send(s, buffer+uploaded, uploadSize, 0);
                    if (charactersWritten>0)
                    {
                        uploaded += charactersWritten;
                        if (charactersWritten != uploadSize && uploaded != bufferSize)
                        {
                            cout << "The whole query wasn't properly uploaded to the server :(" << endl;
                        }
                    }
                }
            }
            cout << "Query uploaded." << endl;
            free(buffer);
        }
        
        
        bool getResult()
        {
            int bufferSize = catalogue->resultRows * catalogue->resultCols * sizeof(int64);
            char *buffer = (char *)malloc(bufferSize);
            int64 *res = (int64 *)malloc (bufferSize);

            int charactersRead = 0;
            
            bzero(buffer, bufferSize);
            
            // we can't download all at once
            // let's split it in sections of 1k
            int downloaded = 0;
            int downloadSize = 1024;
            char *tmpBuf = buffer;
            
            while (downloaded != bufferSize)
            {
                int nextDownload = (bufferSize - downloaded)<downloadSize ? (bufferSize - downloaded) : downloadSize;
            
                charactersRead=recv(s, tmpBuf, nextDownload, 0);
                downloaded += charactersRead;
                tmpBuf += charactersRead; // pointer shifting, so as not to override the just written data in the buffer.
                if (charactersRead != downloadSize && downloaded != bufferSize)
                {
                    cout << "I didn't get everything I expected : " << downloaded << " instead of " << bufferSize << endl;
                    return false;
                }
            }
            
            memcpy(res, buffer, bufferSize);
            result->addResult(new Matrix_int64(catalogue->resultRows, catalogue->resultCols, res));
            
            free(buffer);
            return true;
        }
};



/////////////////////////////////////////////////////////
//      entry point to the program
/////////////////////////////////////////////////////////
int main (int argc, char **argv)
{
    srand(time(NULL));
    RIPClient *cli = new RIPClient();
    
    cout << "All done, exiting client." << endl;
    delete(cli);
}




